/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file send.h Sender-related defines and structures. */

/** The sender subcommands. */
enum {SENDER_ADD, SENDER_DELETE, SENDER_ALLOW, SENDER_DENY, SENDER_ON, SENDER_OFF, NUM_SENDER_CMDS};

/**
 * Describes one supported sender of para_server.
 *
 * \sa http_send.c udp_send.c, dccp_send.c.
 */
struct sender {
	/** The name of the sender. */
	const char *name;
	/**
	 * The init function of this sender.
	 *
	 * It must fill in all function pointers of \a s as well as the \a
	 * client_cmds array, see below. It should also do all necessary
	 * preparations to init this sending facility, for example it could
	 * open a tcp port.
	 */
	void (*init)(struct sender *s);
	/**
	 * Return the help text of this sender.
	 *
	 * The result must be dynamically allocated and is freed by the caller.
	 */
	char* (*help)(void);
	/**
	 * Return current status info about this sender.
	 *
	 * The result must be dynamically allocated and is freed by the caller.
	 */
	char* (*info)(void);
	/**
	 * The send-hook.
	 *
	 * It gets called whenever para_server is playing and the current
	 * audio format handler indicates that another chunk of data should
	 * be sent now. The two parameters \a current_chunk and \a chunks_sent
	 * only differ if the stream was repositioned by the \a ff or \a jmp
	 * command. Of course, \a buf is a pointer to the chunk of data which
	 * should be sent, and \a len is the length of this buffer.
	*/
	void (*send)(long unsigned current_chunk, long unsigned chunks_sent,
		const char *buf, size_t len, const char *header_buf,
		size_t header_len);
	/**
	 * Add file descriptors to fd_sets.
	 *
	 * The pre_select function of each supported sender is called just before
	 * para_server enters its main select loop. Each sender may add its own
	 * file descriptors to the \a rfds or the \a wfds set.
	 *
	 * If a file descriptor was added, \a max_fileno must be increased by
	 * this function, if necessary.
	 *
	 * \sa select(2).
	 */
	void (*pre_select)(int *max_fileno, fd_set *rfds, fd_set *wfds);
	/**
	 * Handle the file descriptors which are ready for I/O.
	 *
	 * If the pre_select hook added one ore more file descriptors to the
	 * read or write set, this is the hook to check the result and do any
	 * I/O on those descriptors which are ready for reading/writing.
	 */
	void (*post_select)(fd_set *rfds, fd_set *wfds);
	/**
	 * Terminate all connected clients.
	 *
	 * This is called e.g. if the stop command was executed. It should make
	 * the clients aware of the end-of-file condition.
	 */
	void (*shutdown_clients)(void);
	/**
	 * Array of function pointers for the sender subcommands.
	 *
	 * Each sender may implement any subset of the sender commands by
	 * filling in the appropriate function pointer in the array. A \p NULL
	 * pointer means this command is not implemented by this sender.
	 */
	int (*client_cmds[NUM_SENDER_CMDS])(struct sender_command_data*);
};

/** Describes one client, connected to a paraslash sender. */
struct sender_client {
	/** The file descriptor of the client. */
	int fd;
	/** The socket "name" of the client. */
	char *name;
	/** The position of this client in the client list. */
	struct list_head node;
	/** Non-zero if audio file header has been sent. */
	int header_sent;
	/** The list of pending chunks for this client. */
	struct chunk_queue *cq;
	/** Data specific to the particular sender. */
	void *private_data;
};

/** Describes the current status of one paraslash sender. */
struct sender_status {
	/** The file descriptor of the socket this sender is listening on. */
	int listen_fd;
	/** The TCP/DCCP port used by this sender. */
	int port;
	/** The current number of simultaneous connections. */
	int num_clients;
	/** The maximal number of simultaneous connections. */
	int max_clients;
	/** Whether the access control list is a whitelist. */
	int default_deny;
	/** The whitelist/blacklist. */
	struct list_head acl;
	/** The list of connected clients. */
	struct list_head client_list;
};

void shutdown_client(struct sender_client *sc, struct sender_status *ss);
void shutdown_clients(struct sender_status *ss);
void send_chunk(struct sender_client *sc, struct sender_status *ss,
		size_t max_bytes_per_write, long unsigned current_chunk,
		const char *buf, size_t len, const char *header_buf,
		size_t header_len);
void init_sender_status(struct sender_status *ss, char **access_arg, int num_access_args,
	int port, int max_clients, int default_deny);
char *get_sender_info(struct sender_status *ss, char *name);

void generic_com_allow(struct sender_command_data *scd,
		struct sender_status *ss);
void generic_com_deny(struct sender_command_data *scd,
		struct sender_status *ss);
int generic_com_on(struct sender_status *ss, unsigned protocol);
void generic_com_off(struct sender_status *ss);
char *generic_sender_help(void);
struct sender_client *accept_sender_client(struct sender_status *ss);
int send_queued_chunks(int fd, struct chunk_queue *cq,
		size_t max_bytes_per_write);
int parse_fec_url(const char *arg, struct sender_command_data *scd);
