/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/** \file udp_recv.c Paraslash's udp receiver */

#include <dirent.h>
#include <sys/socket.h>
#include <net/if.h>

#include "para.h"
#include "error.h"
#include "portable_io.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "udp_recv.cmdline.h"
#include "audiod.h"
#include "string.h"
#include "net.h"
#include "fd.h"

/** The size of the receiver node buffer. */
#define UDP_RECV_CHUNK_SIZE (128 * 1024)
/**
 * Data specific to the udp receiver.
 *
 * \sa \ref receiver, \ref receiver_node.
 */
struct private_udp_recv_data {
	/** The socket file descriptor. */
	int fd;
};

static void udp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_udp_recv_data *purd = rn->private_data;

	para_fd_set(purd->fd, &s->rfds, &s->max_fileno);
}

static int enough_space(size_t nbytes, size_t loaded)
{
	return nbytes + loaded < UDP_RECV_CHUNK_SIZE;
}

static int add_rn_output(struct receiver_node *rn, char *buf, size_t len)
{
	if (!len)
		return 1;
	if (!enough_space(len, rn->loaded))
		return -E_UDP_OVERRUN;
	memcpy(rn->buf + rn->loaded, buf, len);
	rn->loaded += len;
	return 1;
}

static void udp_recv_post_select(__a_unused struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_udp_recv_data *purd = rn->private_data;
	int ret;
	char tmpbuf[UDP_RECV_CHUNK_SIZE];
	size_t packet_size;

	if (rn->output_error && *rn->output_error < 0) {
		t->error = *rn->output_error;
		return;
	}
	if (!FD_ISSET(purd->fd, &s->rfds))
		return;
	ret = recv_bin_buffer(purd->fd, tmpbuf, UDP_RECV_CHUNK_SIZE);
	if (ret < 0) {
		if (is_errno(ret, EINTR) || is_errno(ret, EAGAIN))
			goto success;
		t->error = ret;
		return;
	}
	t->error = -E_RECV_EOF;
	if (!ret)
		return;
	packet_size = ret;
	if (packet_size >= FEC_EOF_PACKET_LEN)
		if (!memcmp(tmpbuf, FEC_EOF_PACKET, FEC_EOF_PACKET_LEN))
			return;
	t->error = add_rn_output(rn, tmpbuf, packet_size);
	if (t->error < 0)
		return;
success:
	t->error = 1;
}

static void udp_shutdown(void)
{
	return;
}

static void udp_recv_close(struct receiver_node *rn)
{
	struct private_udp_recv_data *purd = rn->private_data;

	if (purd->fd >= 0)
		close(purd->fd);
	free(rn->private_data);
	free(rn->buf);
}

static void *udp_recv_parse_config(int argc, char **argv)
{
	int ret;
	struct udp_recv_args_info *tmp =
		para_calloc(sizeof(struct udp_recv_args_info));

	ret = udp_recv_cmdline_parser(argc, argv, tmp)? -E_UDP_SYNTAX : 1;
	if (ret >= 0)
		return tmp;
	free(tmp);
	return NULL;
}

/*
 * Perform AF-independent joining of multicast receive addresses.
 *
 * \param fd	Bound socket descriptor.
 * \param iface	The receiving multicast interface, or NULL for the default.
 *
 * \return Zero if okay, negative on error.
 */
static int mcast_receiver_setup(int fd, const char *iface)
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);
	int id = iface == NULL ? 0 : if_nametoindex(iface);

	if (getsockname(fd, (struct sockaddr *)&ss, &sslen) < 0)
		goto err;

	if (iface != NULL && id == 0)
		PARA_WARNING_LOG("could not resolve interface %s, using default", iface);

	switch (ss.ss_family) {
	case AF_INET:
		if (IN_MULTICAST(htonl(((struct sockaddr_in *)&ss)->sin_addr.s_addr))) {
#ifdef HAVE_IP_MREQN
			struct ip_mreqn m4;

			m4.imr_address.s_addr	= INADDR_ANY;
			m4.imr_ifindex		= id;
#else
			struct ip_mreq m4;

			m4.imr_interface.s_addr	= INADDR_ANY;
			if (id != 0)
				PARA_ERROR_LOG("Setting IPv4 receiver mcast interface not supported.");

#endif
			m4.imr_multiaddr	= ((struct sockaddr_in *)&ss)->sin_addr;

			if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &m4, sizeof(m4)) < 0)
				break;
		}
		return 0;
	case AF_INET6:
		if (IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)&ss)->sin6_addr)) {
			struct ipv6_mreq m6;

			memset(&m6, 0, sizeof(m6));
			memcpy(&m6.ipv6mr_multiaddr, &((struct sockaddr_in6 *)&ss)->sin6_addr, 16);
			m6.ipv6mr_interface = id;
			if (setsockopt(fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &m6, sizeof(m6)) < 0)
				break;
		}
		return 0;
	default:
		PARA_ERROR_LOG("address family %d not supported", ss.ss_family);
		return -E_ADDRESS_LOOKUP;
	}
err:
	return -ERRNO_TO_PARA_ERROR(errno);
}

static int udp_recv_open(struct receiver_node *rn)
{
	struct private_udp_recv_data *purd;
	struct udp_recv_args_info *c = rn->conf;
	char  *iface = c->iface_given ? c->iface_arg : NULL;
	int ret;

	rn->buf = para_calloc(UDP_RECV_CHUNK_SIZE);
	rn->private_data = para_calloc(sizeof(struct private_udp_recv_data));
	purd = rn->private_data;

	ret = makesock(AF_UNSPEC, IPPROTO_UDP, 1, c->host_arg, c->port_arg);
	if (ret < 0)
		goto err;
	purd->fd = ret;

	ret = mcast_receiver_setup(purd->fd, iface);
	if (ret < 0) {
		close(purd->fd);
		return ret;
	}

	ret = mark_fd_nonblocking(purd->fd);
	if (ret < 0)
		goto err;
	PARA_NOTICE_LOG("receiving from %s:%d, fd=%d\n", c->host_arg,
		c->port_arg, purd->fd);
	return purd->fd;
err:
	free(rn->private_data);
	free(rn->buf);
	return ret;
}

/**
 * The init function of the udp receiver.
 *
 * \param r Pointer to the receiver struct to initialize.
 *
 * Initialize all function pointers of \a r.
 */
void udp_recv_init(struct receiver *r)
{
	struct udp_recv_args_info dummy;

	udp_recv_cmdline_parser_init(&dummy);
	r->shutdown = udp_shutdown;
	r->open = udp_recv_open;
	r->close = udp_recv_close;
	r->pre_select = udp_recv_pre_select;
	r->post_select = udp_recv_post_select;
	r->parse_config = udp_recv_parse_config;
	r->help = (struct ggo_help) {
		.short_help = udp_recv_args_info_help,
		.detailed_help = udp_recv_args_info_detailed_help
	};
	udp_recv_cmdline_parser_free(&dummy);
}
